@@ -1,3 +1,4 @@
+import asyncio
 import base64
 import datetime
 import os
@@ -20,7 +21,6 @@ from injected_utils.shelf_util import MyShelf
 from pinjected import Injected
 from pinjected.di.proxiable import DelegatedVar
 from pinjected.di.util import check_picklable
-from ray_proxy import RemoteInterpreterFactory, Var
 from loguru import logger
 
 
@@ -87,7 +87,6 @@ class ProtocolWrappedDict:
     def items(self):
         return [(self.protocol.decode_key(k), v) for k, v in self.cache.items()]
 
-
 @dataclass
 class CachedFunction:
     """
@@ -219,7 +218,7 @@ class CachedFunction:
         return CachedFunction(shelf, func, cache_serializer, cache_deserializer)
 
     @staticmethod
-    def create_with_proxy(rif: RemoteInterpreterFactory, gen_cache: Callable[[], Dict], func):
+    def create_with_proxy(rif: "RemoteInterpreterFactory", gen_cache: Callable[[], Dict], func):
         env = rif.create(num_cpus=0)
         r_cache = env.put(gen_cache)()
         return CachedFunction(RemoteDict(r_cache), func)
@@ -302,18 +301,24 @@ class KeyEncoder(IKeyEncoder):
             raise e
 
     def get_key(self, *args, **kwargs):
-        # logger.warning(f"calculating key for {args},{kwargs}")
         bound = self.signature.bind(*args, **kwargs)
         bound.apply_defaults()
 
         key = self._encode_key(frozendict(bound.arguments))
-        # logger.warning(f"calculated key for {args},{kwargs} => {encoded}")
-        # logger.warning(f"key_encoder:{self}")
         assert key is not None, f'key is None. {args},{kwargs}'
-        # logger.debug(f"calculated key for {args},{kwargs} => {key}")
         return key
 
 
+@dataclass
+class AsyncLockMap:
+    locks: Dict[Any, Lock] = field(default_factory=dict)
+
+    def get(self,key):
+        if key not in self.locks:
+            self.locks[key] = asyncio.Lock()
+        return self.locks[key]
+
+
 @dataclass
 class AsyncCachedFunction:
     cache: Dict
@@ -327,6 +332,7 @@ class AsyncCachedFunction:
     name: str = field(default=None)
     value_invalidator: Callable[[Any], Union[bool, Awaitable[bool]]] = field(default=lambda x: False)
     _concurrent_validation_count: int = field(default=0)
+    _lock_map:AsyncLockMap = field(default_factory=AsyncLockMap)
 
     def __post_init__(self):
         from loguru import logger
@@ -383,48 +389,49 @@ class AsyncCachedFunction:
     async def __call__(self, *args, **kwargs):
         key = self.key_encoder.get_key(*args, **kwargs)
         invalidated = False
-        if key in self.cache:
-            # logger.debug(f"async cached function hit for {self.name}! {str(args)[:100]}...,{str(kwargs)}...")
-            try:
-                loaded = await self._load_value(key)
+        async with self._lock_map.get(key):
+            if key in self.cache:
+                # logger.debug(f"async cached function hit for {self.name}! {str(args)[:100]}...,{str(kwargs)}...")
                 try:
-                    if invalidated := await self._invalidate_value(loaded):
-                        logger.warning(
-                            f"invalidated a loaded cache for {self.name}! {str(args)[:100]}...,{str(kwargs)}...")
-                    else:
-                        return loaded
-                except Exception as validation_error:
-                    logger.error(f"error while validating the decoded value")
-                    raise validation_error
+                    loaded = await self._load_value(key)
+                    try:
+                        if invalidated := await self._invalidate_value(loaded):
+                            logger.warning(
+                                f"invalidated a loaded cache for {self.name}! {str(args)[:100]}...,{str(kwargs)}...")
+                        else:
+                            return loaded
+                    except Exception as validation_error:
+                        logger.error(f"error while validating the decoded value")
+                        raise validation_error
+                except Exception as e:
+                    logger.warning(f"error decoding cache for {self.name}:{key} {e}")
+                    logger.warning(f"cache: {self.cache}")
+                    logger.warning(f"cache key: {key}")
+            if key in self.cache:
+                logger.warning(f"cache hit but invalidated for {self.name}! {str(args)[:100]}...,{str(kwargs)}...")
+            if not isinstance(self.cache, dict):
+                logger.info(f"cache miss for {self.name}: {str(args)[:100]}...,{str(kwargs)[:100]}... in {self.cache}")
+            try:
+                result = await self.func(*args, **kwargs)  # this still has image_data in args
             except Exception as e:
-                logger.warning(f"error decoding cache for {self.name}:{key} {e}")
-                logger.warning(f"cache: {self.cache}")
-                logger.warning(f"cache key: {key}")
-        if key in self.cache:
-            logger.warning(f"cache hit but invalidated for {self.name}! {str(args)[:100]}...,{str(kwargs)}...")
-        if not isinstance(self.cache, dict):
-            logger.info(f"cache miss for {self.name}: {str(args)[:100]}...,{str(kwargs)[:100]}... in {self.cache}")
-        try:
-            result = await self.func(*args, **kwargs)  # this still has image_data in args
-        except Exception as e:
-            import traceback
-            trb = "\n".join(traceback.format_exception(e))
-            logger.warning(
-                f"error {type(e)}{e} when running impl:{self.func} for cache {self.name}! {str(args)[:100]}...,{str(kwargs)[:100]}... => not saving\n{trb}"
-            )
+                import traceback
+                trb = "\n".join(traceback.format_exception(e))
+                logger.warning(
+                    f"error {type(e)}{e} when running impl:{self.func} for cache {self.name}! {str(args)[:100]}...,{str(kwargs)[:100]}... => not saving\n{trb}"
+                )
 
-            raise e
-        # logger.info(f"obtained result for {self.name}: {str(args)[:100]}...,{str(kwargs)[:100]}...")
-        if cause := await self._invalidate_value(result):
-            raise RuntimeError(
-                f"the calculated result for cache is invalid! :{self.name}! {str(args)[:100]}...,{str(kwargs)[:100]}... \n cause:{cause}")
-        if invalidated:
-            logger.info(
-                f"cache was invalidated and is now valid for {self.name}: {str(args)[:100]}...,{str(kwargs)[:100]}...")
-        await self._write_value(key, result)
-        if not isinstance(self.cache, dict):
-            logger.info(f"cache written for {self.name}: {str(args)[:100]}...,{str(kwargs)[:100]}...")
-        return result
+                raise e
+            # logger.info(f"obtained result for {self.name}: {str(args)[:100]}...,{str(kwargs)[:100]}...")
+            if cause := await self._invalidate_value(result):
+                raise RuntimeError(
+                    f"the calculated result for cache is invalid! :{self.name}! {str(args)[:100]}...,{str(kwargs)[:100]}... \n cause:{cause}")
+            if invalidated:
+                logger.info(
+                    f"cache was invalidated and is now valid for {self.name}: {str(args)[:100]}...,{str(kwargs)[:100]}...")
+            await self._write_value(key, result)
+            if not isinstance(self.cache, dict):
+                logger.info(f"cache written for {self.name}: {str(args)[:100]}...,{str(kwargs)[:100]}...")
+            return result
 
     def keys(self):
         return [self.decode_key(s) for s in self.cache.keys()]
@@ -437,7 +444,7 @@ class AsyncCachedFunction:
 
 @dataclass
 class RemoteDict:
-    src: Var#[Dict]
+    src: "Var"#[Dict]
 
     def __getitem__(self, item):
         return self.src[item].fetch()
