@@ -1,29 +1,29 @@
 import asyncio
 import inspect
-import threading
 from asyncio import Future
 from dataclasses import dataclass, field
 from hashlib import sha256
 from pathlib import Path
+from pprint import pformat
 from typing import Callable, Dict, ParamSpec, Any
 
 import cloudpickle
 import jsonpickle
 from frozendict import frozendict
-from loguru import logger
+from pinjected.di.metadata.bind_metadata import BindMetadata
 from returns.maybe import Some
 from sqlitedict import SqliteDict
 
-# from data_tree.util import Pickled
-
 from injected_utils.cached_function import AsyncCachedFunction, CachedFunction, KeyEncoder, IStringKeyProtocol, \
     IKeyEncoder
-from pinjected import Injected, injected_function, injected
+from pinjected import Injected, injected_function, injected, instance
 from pinjected.decoration import update_if_registered
-from pinjected.di.bindings import BindMetadata
-from pinjected.di.graph import Providable
 from pinjected.di.proxiable import DelegatedVar
 from pinjected.di.util import get_code_location
+from pinjected.providable import Providable
+from pinjected.v2.resolver import AsyncResolver
+
+# from data_tree.util import Pickled
 
 P = ParamSpec('P')
 Hasher = Callable[[P], Any]
@@ -33,7 +33,7 @@ HasherFactory = Callable[[inspect.Signature], Hasher]
 def ensure_injected(i: Providable):
     match i:
         case str():
-            return Injected.by_name(i)
+            return injected(i)
         case Injected():
             return i
         case type():
@@ -50,25 +50,47 @@ def pickled_injected(cache_path, injected: Injected):
     from data_tree.util import Pickled
     injected = ensure_injected(injected)
 
-    def _impl(session):
-        return Pickled(cache_path, lambda: session[injected]).value
+    def _impl(__resolver__:AsyncResolver):
+        return Pickled(cache_path, lambda: __resolver__.to_blocking()[injected], backend=cloudpickle).value
 
-    return Injected.bind(_impl)
+    return Injected.bind(_impl).add_dynamic_dependencies(*injected.complete_dependencies).proxy
 
 
-RUNNING_TASKS = []
+@instance
+@dataclass
+class AsyncRunTracker:
+    logger: "Logger"
+
+    def __post_init__(self):
+        self.running_tasks = []
+        self.bound = self.logger.bind(name="AsyncRunTracker")
+
+    def log_status(self):
+        self.bound.info(f"running tasks: {pformat([i[1] for i in self.running_tasks])}\n num:{len(self.running_tasks)}")
+
+    def register(self, fut: Future, fn, *args, **kwargs):
+        self.bound.info(f"adding {fn} to running tasks")
+        self.log_status()
+        self.running_tasks.append((fut, fn, args, kwargs))
+
+        def done(fut):
+            self.bound.info(f"removing {fn} from running tasks")
+            self.running_tasks.remove((fut, fn, args, kwargs))
+            self.log_status()
+
+        fut.add_done_callback(done)
 
 
 @injected_function
-async def async_run(thread_pool, /, fn, *args, **kwargs):
+async def async_run(thread_pool, AsyncRunTracker, /, fn, *args, **kwargs):
     # Ah, there are 391 tasks submitted
     # logger.info(f"submitting to thread pool {fn} from {threading.current_thread()}")
-    assert threading.current_thread().name == "MainThread", f"current thread is {threading.current_thread()}"
+    # assert threading.current_thread().name == "MainThread", f"current thread is {threading.current_thread()}"
     fut = thread_pool.submit(fn, *args, **kwargs)
-    RUNNING_TASKS.append(fn)
     # logger.info(f"awaiting {fn} to thread pool")
-    res = await asyncio.wrap_future(fut)
-    RUNNING_TASKS.remove(fn)
+    wrapped = asyncio.wrap_future(fut)
+    AsyncRunTracker.register(wrapped, fn, args, kwargs)
+    res = await wrapped
     # logger.info(f"got result from thread pool")
     # logger.info(f"remaining tasks: {RUNNING_TASKS[:1]} num:{len(RUNNING_TASKS)} ")
 
@@ -78,8 +100,11 @@ async def async_run(thread_pool, /, fn, *args, **kwargs):
 @injected_function
 async def run_in_new_thread(fn, *args, **kwargs):
     res = Future()
+    import threading
 
     def run_task():
+        from loguru import logger
+        # holy fuck loguru!
         try:
             logger.info(f"running {fn} in new thread :{threading.current_thread()}")
             data = fn(*args, **kwargs)
@@ -89,6 +114,7 @@ async def run_in_new_thread(fn, *args, **kwargs):
             res.set_exception(e)
 
     thread = threading.Thread(target=run_task)
+    from loguru import logger
     logger.info(f"starting thread {thread} for {fn}")
     thread.start()
     try:
@@ -102,6 +128,7 @@ async def run_in_new_thread(fn, *args, **kwargs):
 
 
 def provide_cached(cache, async_run, func: Callable, additional_key: Injected):
+    from loguru import logger
     async def task(key, *args, **kwargs):
         return await async_run(func, *args, **kwargs)
 
@@ -179,7 +206,7 @@ def to_async(_async_run, func):
 
 def run_async(func: Injected[Callable]):
     func = ensure_injected(func)
-    return to_async.proxy(func).eval()
+    return to_async(func).eval()
 
 
 def en_async_cached(cache: Injected[Dict], *additional_keys: Injected):
@@ -192,6 +219,7 @@ def en_async_cached(cache: Injected[Dict], *additional_keys: Injected):
     additional_key = Injected.mzip(*[ensure_injected(i) for i in additional_keys])
 
     def _impl(func: Injected[Callable]):
+        from loguru import logger
         assert isinstance(func, Injected)
         logger.info(f"en_async_cached called with: {func}")
         return Injected.bind(
@@ -247,6 +275,7 @@ def async_cached(cache: Injected[Dict],
             value_invalidator=Injected.pure(lambda x: False) if value_invalidator is None else value_invalidator,
             key_encoder_factory=key_encoder_factory
         )
+        res.__is_async_function__ = True
 
         return update_if_registered(
             async_func,
@@ -287,7 +316,7 @@ class PicklableSqliteDict:
 
     @classmethod
     def from_path(cls, path):
-        if isinstance(path,str):
+        if isinstance(path, str):
             path = Path(path)
         assert isinstance(path, Path), f"path must be a Path object, but got {path}"
         path.parent.mkdir(parents=True, exist_ok=True)
@@ -334,15 +363,22 @@ class JsonBackedSqliteDict:
         key = jsonpickle.dumps(instance)
         del self.src[key]
 
+    def values(self):
+        return [jsonpickle.loads(v) for v in self.src.values()]
+
+    def keys(self):
+        return [jsonpickle.loads(k) for k in self.src.keys()]
+
 
 @injected
-def sqlite_cache(cache_path: str):
+def sqlite_cache(cache_path: str | Path):
     # TODO wrap this to be serializable for the ray.
     return PicklableSqliteDict.from_path(cache_path)
 
 
 @injected
-def sqlite_dict(cache_path: str):
+def sqlite_dict(cache_path: str | Path):
+    Path(cache_path).parent.mkdir(parents=True, exist_ok=True)
     return JsonBackedSqliteDict.from_path(cache_path)
 
 
@@ -470,6 +506,7 @@ def sync_cached(cache: Injected[Dict],
 
     def provide_blocking_cache(cache, func, _deps, _hasher_factory):
         def new_impl(deps, *arg, **kwargs):
+            from loguru import logger
             # deps are just for cache key purposes
             logger.info(f"new_impl called with: {func.__name__, arg, kwargs}")
             res = func(*arg, **kwargs)
